/*
 *  Copyright (c) 2010 Silicon Graphics, Inc.  All Rights Reserved.
 *  Copyright 2020 Hewlett Packard Enterprise Development LP
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef __GRU_WARS_H__
#define __GRU_WARS_H__
#include "gru_ugly_kabi_hack.h"
#include "gruhandles.h"

#define UV_GR0_HRB_MMR_CONFIG		(is_uv4_hub() ? 0x0600128 : 0xc00128)
#define UV_GR1_HRB_MMR_CONFIG		(is_uv4_hub() ? 0x0700128 : 0x1000128)
#define UV_GRU_MRU_KEY			0x2a364516	/* arbitrary number */

#define GRU_MRU_BASE			0x30000

#define GENERIC_HSEL			0x6

/* MMR used to set MRU key */
struct uv_hrb_mmr_config {
	union {
		struct {
			int enable;
			int key;
		};
		unsigned long v;
	};
};


enum gr_cbr_fsm_state {
	CB_INACT        = 0x0,
	CB_IDLE         = 0x1,
	CB_PE_CHK       = 0x2,
	CB_DSR_PRE      = 0x3,
	CB_QUEUED       = 0x4,
	CB_ISSUED       = 0x5,
	CB_WT_RSP       = 0x6,
	CB_DSR_POST     = 0x7,
	CB_BSY_INT_FMM  = 0x8,
	CB_INT_FMM      = 0x9,
	CB_BSY_INT_UPM  = 0xa,
	CB_INT_UPM      = 0xb,
	CB_BSY_INT      = 0xc,
	CB_INT          = 0xd,
	CB_BCPY_CHK     = 0xe
};

enum gr_gsh_opc {
        GSH_RD_CB_STATE         = 0x01,         // Read CBR/CBE state starting at index.
        GSH_RD_CB_COMPRESSED    = 0x02,         // Read Compressed CB state (all 128).
        GSH_RD_CCH_STATE        = 0x03,         // Read CCH (defined by index) and current TGH state.
        GSH_RD_CCH_COMPRESSED   = 0x04,         // Read Compressed CCH state (all 16).
        GSH_RD_GRB_COMPRESSED   = 0x05,         // Read Compressed GRB state (all 128).
        GSH_RD_HOLD_STATE       = 0x06,         // Read WBEXC|WBSHRD Hold state (CBR DW0-1, TFH 2-3, CCH 4, TGH 5)

        GSH_GEN_CB_ERROR        = 0x10,         // Generate CB Error
        GSH_GEN_CB_INS_DN       = 0x11,         // Generate CB Instruction Done
        GSH_GEN_CB_ABRT_DN      = 0x12,         // Generate CB Abort Done
        GSH_FORCE_CB_IDLE       = 0x13,         // Force CB to idle

        GSH_WR_CB_STATE         = 0x14,         // Write CB state
        GSH_WR_CB_ALLREP        = 0x15,         // Write CB all reply counters
        GSH_FORCE_CCH_ERROR     = 0x16,         // Force CCH Error
        GSH_FORCE_TGH_INT_DN    = 0x17,         // Force TGH Interrupt Done
        GSH_FORCE_TGH_RSTRT_DN  = 0x18,         // Force TGH Restart Done
        GSH_FORCE_TGH_SYNC_DN   = 0x19,         // Force TGH Sync done
        GSH_FORCE_TGH_ERROR     = 0x1a,         // Force TGH Error
        GSH_RESET_GRU           = 0x1b,         // Basically reset CCH, CEs, CBs and TLB
        GSH_FLUSH_HANDLE        = 0x1c,         // Force a flush of the reqested handle.
};

enum gr_cch_fsm_state {
	CCH_INACTIVE		= 0x0,
	CCH_BSY_ALLOC_RO	= 0x1,
	CCH_BSY_ALLOC		= 0x2,
	CCH_MAPPED		= 0x3,
	CCH_ACT_CBESYNC		= 0x4,
	CCH_ACT_CBRSYNC		= 0x5,
	CCH_ACT_RO		= 0x6,
	CCH_ACT_CBR		= 0x7,
	CCH_ACT_RR		= 0x8,
	CCH_ACTIVE		= 0x9,
	CCH_BSY_INTR		= 0xa,
	CCH_INTERRUPTED		= 0xb,
	CCH_BSY_DSR_SYNC	= 0xc,
	CCH_BSY_TFH_CLN		= 0xd,
	CCH_BSY_DEALLOC		= 0xe,
	CCH_ERROR		= 0xf
};

enum gr_tfh_fsm_state {
	TFS_INACT      	 = 0x0,
	TFS_IDLE       	 = 0x1,
	TFS_MISS_UPM   	 = 0x2,
	TFS_MISS_FMM   	 = 0x3,
	TFS_HW_ERROR   	 = 0x4,
	TFS_WRITE_TLB  	 = 0x5,
	TFS_RESTART_CBR	 = 0x6,
	TFS_WRITO_TLB  	 = 0x7
};

/*
 * GSH CCH State
 */
struct GshCchState {
	unsigned long WhoAmI:4;
	unsigned long cch_state:4;
	unsigned long int_state:3;
	unsigned long cch_active:1;
	unsigned long tgh_int_active:1;
	unsigned long tgh_intr_pend:1;
	unsigned long tgh_rstrt_pend:1;
	unsigned long tgh_sync_pend:1;

	unsigned long tgh_sync_act:1;
	unsigned long cch_int_active:1;
	unsigned long cch_int_pend:1;
	unsigned long cch_sync_pend:1;
	unsigned long inter_sync_act:1;
	unsigned long dealloc_active:1;
	unsigned long gxe_intr_act:1;
	unsigned long cch_map_busy:1;

	unsigned long CCause:4;
	unsigned long CState:2;
	unsigned long CStatus:2;

	unsigned long cch_mapped:1;
	unsigned long cb_idx_busy:1;
	unsigned long dsr_flsh_act:1;
	unsigned long cb_sync_stall:1;
	unsigned long cbr_pending_sync:1;
	unsigned long cbe_sync_wait:1;
	unsigned long cbe_is_unown:1;
	unsigned long cbr_is_unown:1;
	unsigned long cbe_went_unown:1;
	unsigned long tfh_is_inact:1;
	unsigned long cbr_is_inact:1;
	unsigned long cbr_is_intr:1;
	unsigned long tgh_idx:5;
	unsigned long unused:15;
};

/*
 * GSH Control Instruction - cacheline used to check CBR states
 */
union GshCbrFullState {
	struct {
		unsigned long InsState:2;
		unsigned long IntOcc:1;
		unsigned long AbortOcc:1;

		unsigned long InvalOcc:1;
		unsigned long NewFaultActive:1;
		unsigned long FatalExcpOcc:1;
		unsigned long DsrOpPend:1;

		unsigned long CbrState:4;

		unsigned long BcpyActive:1;
		unsigned long AllRepZero:1;
		unsigned long TfsSyncState:2;

		unsigned long TfhState:3;
		unsigned long OldTfhState:3;
		unsigned long Store:1;
		unsigned long Error:1;

		unsigned long Context:4;
		unsigned long ErrorCode:5;
		unsigned long CbInt:1;
		unsigned int fill1:14;
		unsigned long AllRep0:8;
		unsigned long AllRep1:8;
	};
	unsigned long dw;
};

/*
 * GSH Control Instruction - cacheline used to issue GSH instructions to the GRU
 */
struct GshControl {
	unsigned long cmd:1;
	unsigned long busy:1;
	unsigned long error:1;
	unsigned long error2:1;
	unsigned long fill1:4;
	unsigned long opc:5;
	unsigned long fill2:3;

	unsigned long index:7;
	unsigned long fill3:9;

	unsigned long key:32;

	unsigned long value;
};

/* Get the handle number for a valid MCS handle address */
static inline int get_handle_number(void *h)
{
	return (int)(((unsigned long)h & 0x7fff) / GRU_HANDLE_STRIDE);
}


static inline struct GshControl *mcs_to_gsh(void *cch)
{
	return (struct GshControl *)(((unsigned long)cch & ~0xfffff) + GRU_MRU_BASE);
}

static inline void *gsh_to_data0(struct GshControl *gsh)
{
	return (void *)((unsigned long)gsh + 0x200);
}

static inline void *gsh_to_data1(struct GshControl *gsh)
{
	return (void *)((unsigned long)gsh + 0x300);
}

static inline void gsh_flush_cache_line(void *h)
{
	mb();
	gru_flush_cache(h);
	sync_core();
	mb();
}

static inline void do_gsh_command(int blade, struct GshControl *ctl, int opc, int index,
				  unsigned long value)
{
	union {
		struct GshControl ctl;
		unsigned long dw0;
	} tctl;

	tctl.ctl.key = UV_GRU_MRU_KEY;
	tctl.ctl.opc = opc;
	tctl.ctl.index = index;
	tctl.ctl.cmd = 1;
	ctl->value = value;
	mb();
	*(unsigned long *)ctl = tctl.dw0;
	while (ctl->cmd || ctl->busy) {
		cpu_relax();
		gsh_flush_cache_line(ctl);
	}
	if (ctl->error)
		panic("GRU: do_gsh_command - failure. blade %d, gsh %p, opc 0x%x, index 0x%d, cmd %d, busy %d, error %d, error2 %d\n",
		      blade, ctl, ctl->opc, ctl->index, ctl->cmd, ctl->busy, ctl->error, ctl->error2);
}

#endif /* __GRU_WARS_H__ */
