/*
 *  Copyright (c) 2008 Silicon Graphics, Inc.  All Rights Reserved.
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation; either version 2.1 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA
 */

#ifndef __GRULIB_H__
#define __GRULIB_H__
#include "gru_ugly_kabi_hack.h"

#define GRU_BASENAME		"gru"
#define GRU_FULLNAME		"/dev/gru"
#define GRU_IOCTL_NUM 		 'G'
#define TASK_COMM_LEN		16

/*
 * Maximum number of GRU segments that a user can have open
 */
#define GRU_MAX_OPEN_CONTEXTS		32

/* Set Number of Request Blocks */
#define GRU_CREATE_CONTEXT		_IOWR(GRU_IOCTL_NUM, 1, void *)

/*  Set Context Options */
#define GRU_SET_CONTEXT_OPTION		_IOWR(GRU_IOCTL_NUM, 4, void *)

/* Fetch exception detail */
#define GRU_USER_GET_EXCEPTION_DETAIL	_IOWR(GRU_IOCTL_NUM, 6, void *)

/* For user call_os handling with tracing - normally a TLB fault */
#define GRU_USER_CALL_OS_TRACE		_IOWR(GRU_IOCTL_NUM, 7, void *)

/* For user call_os handling - normally a TLB fault */
#define GRU_USER_CALL_OS		_IOWR(GRU_IOCTL_NUM, 8, void *)

/* For user unload context */
#define GRU_USER_UNLOAD_CONTEXT		_IOWR(GRU_IOCTL_NUM, 9, void *)

/* For dumping GRU chiplet state */
#define GRU_DUMP_CHIPLET_STATE		_IOWR(GRU_IOCTL_NUM, 11, void *)

/* For getting gseg statistics */
#define GRU_GET_GSEG_STATISTICS		_IOWR(GRU_IOCTL_NUM, 12, void *)

/* For switching a GRU to FMM polling mode */
#define GRU_FMM_POLLING_MODE		_IOWR(GRU_IOCTL_NUM, 13, void *)

/* For user TLB flushing (primarily for tests) */
#define GRU_USER_FLUSH_TLB		_IOWR(GRU_IOCTL_NUM, 50, void *)

/* Various kernel self-tests */
#define GRU_KTEST			_IOWR(GRU_IOCTL_NUM, 52, void *)

/* Increment message queue statistic (TEMP) */
#define GRU_MESSAGE_QUEUE_STATISTIC	_IOWR(GRU_IOCTL_NUM, 53, void *)
#define ENCODE_MQSTAT(n, v)		((void *)((unsigned long)(v) << 8 | (n)))
#define MQSTAT_TO_ID(a)			((a) & 255)
#define MQSTAT_TO_VAL(a)		((a) >> 8)

/* for the /proc/sgi_uv/gru/debug_options file */
#define GRU_SET_STATISTICS_ENABLE	_IOWR(GRU_IOCTL_NUM, 54, void *)

/* convert a user vaddr to a physical address */
#define GRU_USER_VTOP			_IOWR(GRU_IOCTL_NUM, 55, void *)

#define CONTEXT_WINDOW_BYTES(th)        (GRU_GSEG_PAGESIZE * (th))
#define THREAD_POINTER(p, th)		(p + GRU_GSEG_PAGESIZE * (th))
#define GSEG_START(cb)			((void *)((unsigned long)(cb) & ~(GRU_GSEG_PAGESIZE - 1)))

/*
 * Size used to map GRU GSeg
 */
#define GRU_GSEG_PAGESIZE    (256 * 1024UL)

/*
 * Exceptions are retried for the following cases. If any OTHER bits are set
 * in ecause, the exception is not retryable.
 */
#define EXCEPTION_RETRY_BITS (CBE_CAUSE_EXECUTION_HW_ERROR |            \
			      CBE_CAUSE_TLBHW_ERROR |                   \
			      CBE_CAUSE_RA_REQUEST_TIMEOUT |            \
			      CBE_CAUSE_RA_RESPONSE_NON_FATAL |         \
			      CBE_CAUSE_HA_RESPONSE_NON_FATAL |         \
			      CBE_CAUSE_RA_RESPONSE_DATA_ERROR |        \
			      CBE_CAUSE_HA_RESPONSE_DATA_ERROR          \
			     )


struct gru_get_gseg_statistics_req {
	unsigned long			gseg;
	struct gru_gseg_statistics	stats;
};

/*
 * Structure used for UPM call OS requests.
 */
struct gru_call_os_req {
	unsigned long	cb;		/* User CB address */
	unsigned long	fault_vaddr;	/* ret: address that caused fault. NULL if none. */
	int		gid;
	int		dropins;
	int		asid;
	unsigned char	pageshift;
	unsigned char	write;
};

/*
 * Structure used to pass TLB flush parameters to the driver
 */
struct gru_create_context_req {
	unsigned long		gseg;
	unsigned int		data_segment_bytes;
	unsigned int		control_blocks;
	unsigned int		maximum_thread_count;
	unsigned int		options;
	int			tlb_preload_count;
};

/*
 * Structure used to pass unload context parameters to the driver
 */
struct gru_unload_context_req {
	unsigned long	gseg;
};
/*
 * Structure used to convert a user virtual address to a physical address
 */
struct gru_user_vtop_req {
	unsigned long	gseg;
	unsigned long	vaddr;
	unsigned long	paddr;
	int		pageshift;
};

/*
 * Structure used to initiate GRU polling for TLB misses
 */
struct gru_fmm_polling_req {
	unsigned int	gid;
};

/*
 * Structure used to set context options
 */
enum {sco_gseg_owner, sco_cch_req_slice, sco_blade_chiplet, sco_tlb_preload_pages};
struct gru_set_context_option_req {
	unsigned long	gseg;
	int		op;
	int		val0;
	long		val1;
};

/*
 * Structure used to pass TLB flush parameters to the driver
 */
struct gru_flush_tlb_req {
	unsigned long	gseg;
	unsigned long	vaddr;
	size_t		len;
};

/*
 * Structure used to pass TLB flush parameters to the driver
 */
enum {dcs_pid, dcs_gid};
enum {do_none = 0, do_dsr = 1, do_tlb = 2, do_cbrs = 4};

struct tlb_entry {
	unsigned long	vpn;
	unsigned long	pfn;
	unsigned int	asid;
	unsigned char	valid;
	unsigned char	larger;
	unsigned char	dirty;
	unsigned char	gaa;
};

struct gru_dump_chiplet_state_req {
	unsigned int	op;
	unsigned int	gid;
	int		ctxnum;
	char		data_opt;
	char		lock_cch;
	char		flush_cbrs;
	char		fill[10];
	pid_t		pid;
	void		*buf;
	size_t		buflen;
	/* ---- output --- */
	unsigned int	num_contexts;
	unsigned long	gsh_cb_state[16];
	unsigned long	gsh_cch_state[16];
	/* ---- extended input --- */
	size_t		tlblen;
	struct tlb_entry *tlb;
};

/*
 * Structure used to fetch exception detail for CBs that terminate with
 * CBS_EXCEPTION
 */
struct control_block_extended_exc_detail {
	unsigned long	cb;
	int		opc;
	int		ecause;
	int		exopc;
	long		exceptdet0;
	long		exceptdet1;
	int		cbrstate;
	int		cbrexecstatus;
};

#define GRU_DUMP_MAGIC	0x3474ab6c
struct gru_dump_context_header {
	unsigned int	magic;
	unsigned int	gid;
	unsigned char	ctxnum;
	unsigned char	cbrcnt;
	unsigned char	dsrcnt;
	pid_t		pid;
	char		comm[TASK_COMM_LEN];
	unsigned long	vaddr;
	int		cch_locked;
	unsigned long	data[0];
};

/*-------------------- debugfs interfaces -------------------------------*/

struct tlbdropin_entry {
	unsigned int		asid;
	unsigned int		pid;
	unsigned long		vaddr;
};

struct tlbdropin_info {
	unsigned char		overflow;
	unsigned int		entries;
	struct tlbdropin_entry	ent[0];
};
#endif /* __GRULIB_H__ */
